@@ -95,12 +95,18 @@ def mocked_http_conn(*args, **kwargs):
         yield fake_conn
 
 
+@contextmanager
+def noop_timeout(duration):
+    yield
+
+
 @patch_policies
 class TestDirectClient(unittest.TestCase):
 
     def setUp(self):
-        self.node = {'ip': '1.2.3.4', 'port': '6200', 'device': 'sda',
-                     'replication_ip': '1.2.3.5', 'replication_port': '7000'}
+        self.node = json.loads(json.dumps({  # json roundtrip to ring-like
+            'ip': '1.2.3.4', 'port': '6200', 'device': 'sda',
+            'replication_ip': '1.2.3.5', 'replication_port': '7000'}))
         self.part = '0'
 
         self.account = u'\u062a account'
@@ -116,6 +122,10 @@ class TestDirectClient(unittest.TestCase):
                 self.account, self.container, self.obj))
         self.user_agent = 'direct-client %s' % os.getpid()
 
+        patcher = mock.patch.object(direct_client, 'Timeout', noop_timeout)
+        patcher.start()
+        self.addCleanup(patcher.stop)
+
     def test_gen_headers(self):
         stub_user_agent = 'direct-client %s' % os.getpid()
 
@@ -269,12 +279,17 @@ class TestDirectClient(unittest.TestCase):
             fake_connect.return_value.getresponse.return_value.status = 200
             direct_client.direct_delete_account(self.node, part, account)
             args, kwargs = fake_connect.call_args
+            ip = args[0]
+            self.assertEqual(self.node['ip'], ip)
+            port = args[1]
+            self.assertEqual(self.node['port'], port)
             method = args[2]
             self.assertEqual('DELETE', method)
             path = args[3]
             self.assertEqual('/sda/0/a', path)
             headers = args[4]
-            self.assertTrue('X-Timestamp' in headers)
+            self.assertIn('X-Timestamp', headers)
+            self.assertIn('User-Agent', headers)
 
     def test_direct_delete_account_failure(self):
         part = '0'
@@ -285,8 +300,12 @@ class TestDirectClient(unittest.TestCase):
                 direct_client.direct_delete_account(self.node, part, account)
             except ClientException as err:
                 pass
+            self.assertEqual(self.node['ip'], conn.host)
+            self.assertEqual(self.node['port'], conn.port)
             self.assertEqual('DELETE', conn.method)
             self.assertEqual('/sda/0/a', conn.path)
+            self.assertIn('X-Timestamp', conn.req_headers)
+            self.assertIn('User-Agent', conn.req_headers)
             self.assertEqual(err.http_status, 500)
 
     def test_direct_head_container(self):
@@ -295,6 +314,8 @@ class TestDirectClient(unittest.TestCase):
         with mocked_http_conn(200, headers) as conn:
             resp = direct_client.direct_head_container(
                 self.node, self.part, self.account, self.container)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'HEAD')
             self.assertEqual(conn.path, self.container_path)
 
@@ -314,6 +335,8 @@ class TestDirectClient(unittest.TestCase):
             else:
                 self.fail('ClientException not raised')
             # check request
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'HEAD')
             self.assertEqual(conn.path, self.container_path)
 
@@ -323,7 +346,7 @@ class TestDirectClient(unittest.TestCase):
         self.assertTrue('HEAD' in str(err))
 
     def test_direct_head_container_deleted(self):
-        important_timestamp = Timestamp(time.time()).internal
+        important_timestamp = Timestamp.now().internal
         headers = HeaderKeyDict({'X-Backend-Important-Timestamp':
                                  important_timestamp})
 
@@ -335,6 +358,8 @@ class TestDirectClient(unittest.TestCase):
                 self.assertTrue(isinstance(err, ClientException))
             else:
                 self.fail('ClientException not raised')
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'HEAD')
             self.assertEqual(conn.path, self.container_path)
 
@@ -388,8 +413,7 @@ class TestDirectClient(unittest.TestCase):
             resp_headers, resp = direct_client.direct_get_container(
                 self.node, self.part, self.account, self.container)
 
-        self.assertEqual(conn.req_headers['user-agent'],
-                         'direct-client %s' % os.getpid())
+        self.assertEqual(conn.req_headers['user-agent'], self.user_agent)
         self.assertEqual(headers, resp_headers)
         self.assertEqual([], resp)
 
@@ -397,6 +421,8 @@ class TestDirectClient(unittest.TestCase):
         with mocked_http_conn(200) as conn:
             direct_client.direct_delete_container(
                 self.node, self.part, self.account, self.container)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.container_path)
 
@@ -408,6 +434,8 @@ class TestDirectClient(unittest.TestCase):
             direct_client.direct_delete_container(
                 self.node, self.part, self.account, self.container,
                 headers=headers)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.container_path)
             self.assertTrue('X-Timestamp' in conn.req_headers)
@@ -423,12 +451,75 @@ class TestDirectClient(unittest.TestCase):
             else:
                 self.fail('ClientException not raised')
 
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.container_path)
 
         self.assertEqual(err.http_status, 500)
         self.assertTrue('DELETE' in str(err))
 
+    def test_direct_put_container(self):
+        body = 'Let us begin with a quick introduction'
+        headers = {'x-foo': 'bar', 'Content-Length': str(len(body)),
+                   'Content-Type': 'application/json',
+                   'User-Agent': 'my UA'}
+
+        with mocked_http_conn(204) as conn:
+            rv = direct_client.direct_put_container(
+                self.node, self.part, self.account, self.container,
+                contents=body, headers=headers)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
+            self.assertEqual(conn.method, 'PUT')
+            self.assertEqual(conn.path, self.container_path)
+            self.assertEqual(conn.req_headers['Content-Length'],
+                             str(len(body)))
+            self.assertEqual(conn.req_headers['Content-Type'],
+                             'application/json')
+            self.assertEqual(conn.req_headers['User-Agent'], 'my UA')
+            self.assertTrue('x-timestamp' in conn.req_headers)
+            self.assertEqual('bar', conn.req_headers.get('x-foo'))
+            self.assertEqual(md5(body).hexdigest(), conn.etag.hexdigest())
+        self.assertIsNone(rv)
+
+    def test_direct_put_container_chunked(self):
+        body = 'Let us begin with a quick introduction'
+        headers = {'x-foo': 'bar', 'Content-Type': 'application/json'}
+
+        with mocked_http_conn(204) as conn:
+            rv = direct_client.direct_put_container(
+                self.node, self.part, self.account, self.container,
+                contents=body, headers=headers)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
+            self.assertEqual(conn.method, 'PUT')
+            self.assertEqual(conn.path, self.container_path)
+            self.assertEqual(conn.req_headers['Transfer-Encoding'], 'chunked')
+            self.assertEqual(conn.req_headers['Content-Type'],
+                             'application/json')
+            self.assertTrue('x-timestamp' in conn.req_headers)
+            self.assertEqual('bar', conn.req_headers.get('x-foo'))
+            self.assertNotIn('Content-Length', conn.req_headers)
+            expected_sent = '%0x\r\n%s\r\n0\r\n\r\n' % (len(body), body)
+            self.assertEqual(md5(expected_sent).hexdigest(),
+                             conn.etag.hexdigest())
+        self.assertIsNone(rv)
+
+    def test_direct_put_container_fail(self):
+        with mock.patch('swift.common.bufferedhttp.http_connect_raw',
+                        side_effect=Exception('conn failed')):
+            with self.assertRaises(Exception) as cm:
+                direct_client.direct_put_container(
+                    self.node, self.part, self.account, self.container)
+        self.assertEqual('conn failed', str(cm.exception))
+
+        with mocked_http_conn(Exception('resp failed')):
+            with self.assertRaises(Exception) as cm:
+                direct_client.direct_put_container(
+                    self.node, self.part, self.account, self.container)
+        self.assertEqual('resp failed', str(cm.exception))
+
     def test_direct_put_container_object(self):
         headers = {'x-foo': 'bar'}
 
@@ -436,12 +527,14 @@ class TestDirectClient(unittest.TestCase):
             rv = direct_client.direct_put_container_object(
                 self.node, self.part, self.account, self.container, self.obj,
                 headers=headers)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'PUT')
             self.assertEqual(conn.path, self.obj_path)
             self.assertTrue('x-timestamp' in conn.req_headers)
             self.assertEqual('bar', conn.req_headers.get('x-foo'))
 
-        self.assertEqual(rv, None)
+        self.assertIsNone(rv)
 
     def test_direct_put_container_object_error(self):
         with mocked_http_conn(500) as conn:
@@ -454,6 +547,8 @@ class TestDirectClient(unittest.TestCase):
             else:
                 self.fail('ClientException not raised')
 
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'PUT')
             self.assertEqual(conn.path, self.obj_path)
 
@@ -464,10 +559,12 @@ class TestDirectClient(unittest.TestCase):
         with mocked_http_conn(204) as conn:
             rv = direct_client.direct_delete_container_object(
                 self.node, self.part, self.account, self.container, self.obj)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.obj_path)
 
-        self.assertEqual(rv, None)
+        self.assertIsNone(rv)
 
     def test_direct_delete_container_obj_error(self):
         with mocked_http_conn(500) as conn:
@@ -480,6 +577,8 @@ class TestDirectClient(unittest.TestCase):
             else:
                 self.fail('ClientException not raised')
 
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.obj_path)
 
@@ -493,13 +592,14 @@ class TestDirectClient(unittest.TestCase):
             resp = direct_client.direct_head_object(
                 self.node, self.part, self.account, self.container,
                 self.obj, headers=headers)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'HEAD')
             self.assertEqual(conn.path, self.obj_path)
 
         self.assertEqual(conn.req_headers['user-agent'], self.user_agent)
         self.assertEqual('bar', conn.req_headers.get('x-foo'))
-        self.assertTrue('x-timestamp' not in conn.req_headers,
-                        'x-timestamp was in HEAD request headers')
+        self.assertNotIn('x-timestamp', conn.req_headers)
         self.assertEqual(headers, resp)
 
     def test_direct_head_object_error(self):
@@ -512,6 +612,8 @@ class TestDirectClient(unittest.TestCase):
                 pass
             else:
                 self.fail('ClientException not raised')
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'HEAD')
             self.assertEqual(conn.path, self.obj_path)
 
@@ -519,7 +621,7 @@ class TestDirectClient(unittest.TestCase):
         self.assertTrue('HEAD' in str(err))
 
     def test_direct_head_object_not_found(self):
-        important_timestamp = Timestamp(time.time()).internal
+        important_timestamp = Timestamp.now().internal
         stub_headers = {'X-Backend-Important-Timestamp': important_timestamp}
         with mocked_http_conn(404, headers=stub_headers) as conn:
             try:
@@ -530,6 +632,8 @@ class TestDirectClient(unittest.TestCase):
                 pass
             else:
                 self.fail('ClientException not raised')
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'HEAD')
             self.assertEqual(conn.path, self.obj_path)
 
@@ -543,6 +647,8 @@ class TestDirectClient(unittest.TestCase):
         with mocked_http_conn(200, body=contents) as conn:
             resp_header, obj_body = direct_client.direct_get_object(
                 self.node, self.part, self.account, self.container, self.obj)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'GET')
             self.assertEqual(conn.path, self.obj_path)
         self.assertEqual(obj_body, contents.getvalue())
@@ -557,6 +663,8 @@ class TestDirectClient(unittest.TestCase):
                 pass
             else:
                 self.fail('ClientException not raised')
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'GET')
             self.assertEqual(conn.path, self.obj_path)
 
@@ -577,6 +685,8 @@ class TestDirectClient(unittest.TestCase):
                 except StopIteration:
                     break
                 downloaded += chunk
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual('GET', conn.method)
             self.assertEqual(self.obj_path, conn.path)
         self.assertEqual('123456', downloaded)
@@ -590,6 +700,8 @@ class TestDirectClient(unittest.TestCase):
             direct_client.direct_post_object(
                 self.node, self.part, self.account, self.container, self.obj,
                 headers)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'POST')
             self.assertEqual(conn.path, self.obj_path)
 
@@ -608,6 +720,8 @@ class TestDirectClient(unittest.TestCase):
                 pass
             else:
                 self.fail('ClientException not raised')
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'POST')
             self.assertEqual(conn.path, self.obj_path)
             for header in headers:
@@ -622,9 +736,11 @@ class TestDirectClient(unittest.TestCase):
         with mocked_http_conn(200) as conn:
             resp = direct_client.direct_delete_object(
                 self.node, self.part, self.account, self.container, self.obj)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.obj_path)
-        self.assertEqual(resp, None)
+        self.assertIsNone(resp)
 
     def test_direct_delete_object_with_timestamp(self):
         # ensure timestamp is different from any that might be auto-generated
@@ -634,6 +750,8 @@ class TestDirectClient(unittest.TestCase):
             direct_client.direct_delete_object(
                 self.node, self.part, self.account, self.container, self.obj,
                 headers=headers)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.obj_path)
             self.assertTrue('X-Timestamp' in conn.req_headers)
@@ -662,8 +780,8 @@ class TestDirectClient(unittest.TestCase):
                                                           self.part, ['a83'])
             self.assertEqual(conn.method, 'REPLICATE')
             self.assertEqual(conn.path, '/sda/0/a83')
-            self.assertEqual(conn.host, '1.2.3.5')
-            self.assertEqual(conn.port, '7000')
+            self.assertEqual(conn.host, self.node['replication_ip'])
+            self.assertEqual(conn.port, self.node['replication_port'])
             self.assertEqual(data, resp)
 
     def _test_direct_get_suffix_hashes_fail(self, status_code):
@@ -695,6 +813,8 @@ class TestDirectClient(unittest.TestCase):
             resp = direct_client.direct_put_object(
                 self.node, self.part, self.account, self.container, self.obj,
                 contents, 6)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'PUT')
             self.assertEqual(conn.path, self.obj_path)
         self.assertEqual(md5('123456').hexdigest(), resp)
@@ -711,6 +831,8 @@ class TestDirectClient(unittest.TestCase):
                 pass
             else:
                 self.fail('ClientException not raised')
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'PUT')
             self.assertEqual(conn.path, self.obj_path)
         self.assertEqual(err.http_status, 500)
@@ -722,6 +844,8 @@ class TestDirectClient(unittest.TestCase):
             resp = direct_client.direct_put_object(
                 self.node, self.part, self.account, self.container, self.obj,
                 contents)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual(conn.method, 'PUT')
             self.assertEqual(conn.path, self.obj_path)
         self.assertEqual(md5('6\r\n123456\r\n0\r\n\r\n').hexdigest(), resp)
@@ -733,6 +857,8 @@ class TestDirectClient(unittest.TestCase):
             resp = direct_client.direct_put_object(
                 self.node, self.part, self.account, self.container, self.obj,
                 contents, etag="testing-etag", content_type='Text')
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual('PUT', conn.method)
             self.assertEqual(self.obj_path, conn.path)
             self.assertEqual(conn.req_headers['Content-Length'], '0')
@@ -748,6 +874,8 @@ class TestDirectClient(unittest.TestCase):
             resp = direct_client.direct_put_object(
                 self.node, self.part, self.account, self.container, self.obj,
                 contents, headers=stub_headers)
+            self.assertEqual(conn.host, self.node['ip'])
+            self.assertEqual(conn.port, self.node['port'])
             self.assertEqual('PUT', conn.method)
             self.assertEqual(conn.req_headers['Content-length'], '6')
         self.assertEqual(md5('123456').hexdigest(), resp)
@@ -779,7 +907,8 @@ class TestDirectClient(unittest.TestCase):
         self.assertEqual(err_ctx.exception.http_status, 500)
         self.assertIn('DELETE', err_ctx.exception.message)
         self.assertIn(quote('/%s/%s/%s/%s/%s'
-                            % (self.node['device'], self.part, self.account,
+                            % (self.node['device'].encode('utf-8'),
+                               self.part, self.account,
                                self.container, self.obj)),
                       err_ctx.exception.message)
         self.assertIn(self.node['ip'], err_ctx.exception.message)
@@ -814,5 +943,14 @@ class TestDirectClient(unittest.TestCase):
         for line in error_lines:
             self.assertIn('Kaboom!', line)
 
+
+class TestUTF8DirectClient(TestDirectClient):
+
+    def setUp(self):
+        super(TestUTF8DirectClient, self).setUp()
+        self.account = self.account.encode('utf-8')
+        self.container = self.container.encode('utf-8')
+        self.obj = self.obj.encode('utf-8')
+
 if __name__ == '__main__':
     unittest.main()
